from agenticos.connectors import BaseWorkflowConfig, CrewaiWorkflowConfig
from agenticos.node.models import AgenticConfig

workflows : dict[str,BaseWorkflowConfig] = {}

# Example workflow
# Import the Crew class. If you used the flow from CrewAI docs the following import should work
# If you are getting any erros please correct the import path
from {{folder_name}}.crew import {{class_name}}Crew as Crew

# Define workflows here
# key is the workflow name
workflows["crewai"] = CrewaiWorkflowConfig(
    # Description of the workflow
    description="A crew for running AI tasks",
    # Inputs for the workflow
    inputs={"topic": "The topic to joke about"},
    # The crew class to be used
    crew_cls=Crew,
)

# Create the config object. Change the name to the name of your node
config = AgenticConfig(name="Test Node", workflows=workflows)



import datetime
import json
import logging
import time
from threading import Event, Lock, Thread
from uuid import UUID

import websocket

from agenticos.connectors import BaseWorkflowConfig, BaseWorkflowRunner
from agenticos.node.http_support import HTTPHealthServer
from agenticos.node.models import *
from agenticos.node.settings import settings

log = logging.getLogger(__name__)
jobs: dict[UUID, Job] = {}

ID_FILE = ".node_id"


class RepeatTimer(Thread):
    def __init__(self, event, callback, frequency=30):
        Thread.__init__(self)
        self._stopped = event
        self._callback = callback
        self._frequency = frequency

    def run(self):
        while True:
            self._callback()
            if self._stopped.wait(self._frequency):
                print("@stopping@")
                break


class WSNode:
    def __init__(self, registry: str, config: AgenticConfig):
        self.registry = registry
        self.config = config
        print("Config", config.model_dump())
        self.lock = Lock()
        self.id = None

    def run_job(self, workflow: BaseWorkflowConfig, job: Job) -> None:
        wr = workflow.new_runner()
        job.output = None
        steps = [False] * len(workflow.steps_description())

        def follow_workflow():
            nonlocal workflow, self
            for step_idx, step_output in enumerate(wr.step_output()):
                if steps[step_idx]:
                    continue
                steps[step_idx] = True
                job.output = step_output
                step_finished_msg = StepFinishedMessage(
                    job_id=str(job.id),
                    step=step_idx,
                    result=step_output,
                )
                self.send_ws_message(json.dumps(step_finished_msg.model_dump()))

        try:
            stopFlag = Event()
            thread = RepeatTimer(stopFlag, follow_workflow, 1)
            thread.start()
            time.sleep(0.5)
            wr.start(inputs=job.inputs)
            job.status = JobStatus.COMPLETED
        except Exception as e:
            job.output = str(e)
            job.status = JobStatus.FAILED
        finally:
            time.sleep(1)
            stopFlag.set()

    def on_message(self, ws, msg):
        message = json.loads(msg)
        if "type" not in message:
            log.warning("Unknown message format", message)
        if message["type"] == MSG_HS_ACK:
            log.info("Handshake successful")
            self.id = message["id"]
            # Write the id to ID_FILE
            with open(ID_FILE, "w") as f:
                f.write(self.id)
        elif message["type"] == MSG_JOB_REQ:
            log.debug("Job request: ", message)
            t_req = AgenticJobRequestMessage(**message)
            job = Job(
                id=t_req.job.job_id,
                inputs=t_req.job.inputs,
                status=JobStatus.RUNNING,
                output="",
            )
            jobs[t_req.job.job_id] = job
            workflow = self.config.workflows[t_req.job.workflow]
            thread = Thread(target=self.run_and_report, args=(workflow, job))
            thread.start()

    def run_and_report(self, workflow: BaseWorkflowConfig, job: Job):
        log.debug("Running job", job.id)
        self.run_job(workflow, job)
        tf_msg = JobFinishedMessage(
            job_id=str(job.id), status=job.status, result=job.output
        )
        time.sleep(1)
        log.debug("Sending job finished message", tf_msg.model_dump())
        self.send_ws_message(json.dumps(tf_msg.model_dump()))

    def send_ws_message(self, message):
        # Make sure that only one thread is sending messages at a time
        with self.lock:
            log.debug(
                f"Sending message: {message}",
            )
            self.ws.send(message)

    def on_error(self, ws, error):
        log.error(error)

    def on_close(self, ws, close_status_code, close_msg):
        self.stopFlag.set()
        if hasattr(self, "health_http_server"):
            self.health_http_server.stop()
        log.info("### closed ###")

    def on_open(self, ws):
        # if ID_FILE exists, read the id from it
        try:
            with open(ID_FILE, "r") as f:
                self.id = f.read()
                self.config.id = self.id
        except FileNotFoundError:
            pass
        if settings.HTTP_HEALTHCHECK:
            self.health_http_server = HTTPHealthServer(settings.HTTP_PORT)
            self.health_http_server.start()
        payload = json.dumps(self.config.model_dump())
        ws.send(payload)
        self._init_heartbeat()

    def send_heartbeat(self):
        log.debug("Sending heartbeat")
        self.send_ws_message(self.hearbeat_msg)

    def _init_heartbeat(self):
        log.debug("Init heartbeat")
        self.hearbeat_msg = json.dumps(AgenticMessage(type=MSG_HEARTBEAT).model_dump())
        self.stopFlag = Event()
        thread = RepeatTimer(self.stopFlag, self.send_heartbeat)
        thread.start()
        # this will stop the timer

    def connect_to_registry(self) -> None:
        print(log.getEffectiveLevel())
        if log.getEffectiveLevel() <= 10:  # 10 is DEBUG
            websocket.enableTrace(True)
        hdrs = []
        if settings.AUTH_TOKEN != "":
            hdrs.append("Authorization:Bearer " + settings.AUTH_TOKEN)
        self.ws = websocket.WebSocketApp(
            self.registry + "/ws/nodes/connect",
            on_open=self.on_open,
            on_message=self.on_message,
            on_error=self.on_error,
            on_close=self.on_close,
            header=hdrs,
        )
        self.ws.run_forever()



import logging

from pydantic_settings import BaseSettings

HTTPSERVER = "httpserver"
REGISTRY = "registry"


class Settings(BaseSettings):
    AUTH_TOKEN: str = ""
    REGISTRY_URL: str = "ws://localhost:8080"
    NODE_MODE: str = HTTPSERVER
    HTTP_PORT: int = 8000
    AGENTIC_CONFIG_PATH: str = "src/agentic/agentic_config.py"
    HTTP_HEALTHCHECK: bool = False
    LOG_LEVEL: str = "INFO"
    BASIC_AUTH_USERNAME: str = ""
    BASIC_AUTH_PASSWORD: str = ""

    def registry_rest_url(self) -> str:
        proto, host = self.REGISTRY_URL.split("://")
        if proto == "ws":
            proto = "http"
        elif proto == "wss":
            proto = "https"
        return f"{proto}://{host}"

    def websocket_rest_url(self) -> str:
        proto, host = self.REGISTRY_URL.split("://")
        if proto == "http":
            proto = "ws"
        elif proto == "https":
            proto = "wss"
        return f"{proto}://{host}"
    
    def has_basic_auth(self) -> bool:
        return self.BASIC_AUTH_USERNAME != "" and self.BASIC_AUTH_PASSWORD != ""

settings = Settings()

CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
WARNING = 30
WARN = WARNING
INFO = 20
DEBUG = 10
NOTSET = 0

_nameToLevel = {
    "CRITICAL": CRITICAL,
    "FATAL": FATAL,
    "ERROR": ERROR,
    "WARN": WARNING,
    "WARNING": WARNING,
    "INFO": INFO,
    "DEBUG": DEBUG,
    "NOTSET": NOTSET,
}

_env_level = settings.LOG_LEVEL
_log_level = logging.INFO
if _env_level not in _nameToLevel:
    print(f"Invalid log level: {_env_level}. Defaulting to INFO")
else:
    _log_level = _nameToLevel[_env_level]
print("Setting log level to: ", _log_level)

logging.basicConfig(
    level=_log_level,
    format="%(asctime)s | %(name)s | %(levelname)s | %(message)s",
)



from typing import Dict
from uuid import UUID

from fastapi import BackgroundTasks, FastAPI, HTTPException
from fastapi.responses import JSONResponse

from agenticos.connectors import BaseWorkflowConfig
from agenticos.node.models import AgenticConfig, Job, JobStatus, Workflow

jobs: dict[UUID, Job] = {}
pw: dict[str, Workflow] = {}
workflows: dict[str, BaseWorkflowConfig] = {}

app = FastAPI()


def run_job(workflow: BaseWorkflowConfig, job: Job) -> None:
    try:
        runner = workflow.new_runner()
        runner.start(job.inputs)
        for step in runner.step_output():
            job.output = step
        job.status = JobStatus.COMPLETED
    except Exception as e:
        job.output = str(e)
        job.status = JobStatus.FAILED


def create_server(config: AgenticConfig) -> FastAPI:
    global pw, workflows
    workflows = config.workflows
    for k, v in workflows.items():
        pw[k] = Workflow(name=k, description=v.description(), inputs=v.inputs())
    return app


@app.get(
    "/node/description",
    summary="Get the description of the node",
    response_model=dict[str, Workflow],
)
def description() -> dict[str, Workflow]:
    return pw


@app.post("/workflow/{workflow_name}/run")
async def run(
    workflow_name: str, inputs: dict[str, str], background_tasks: BackgroundTasks
) -> str:
    if workflow_name not in workflows:
        raise HTTPException(status_code=404, detail="Workflow not found")
    job = Job(inputs=inputs, status=JobStatus.RUNNING, output="")
    jobs[job.id] = job
    background_tasks.add_task(run_job, workflows[workflow_name], job)
    return str(job.id)


@app.get("/job/{job_id}")
def get_job(job_id: str) -> Job:
    if UUID(job_id) not in jobs:
        raise HTTPException(status_code=404, detail="Task not found")
    return jobs[UUID(job_id)]


@app.get("/jobs")
def get_jobs() -> list[str]:
    return [str(tk) for tk in jobs.keys()]


@app.get("/health")
def health() -> JSONResponse:
    return JSONResponse({"status": "ok"})



import json
import logging
import time
from typing import Type
from uuid import UUID

import requests
from crewai_tools import BaseTool
from pydantic import BaseModel, Field
from tenacity import retry, stop_after_attempt, stop_after_delay, wait_fixed

from agenticos.node.models import Job, JobStatus, Workflow
from agenticos.node.settings import settings

log = logging.getLogger(__name__)
# Round robin counter for selecting the host
host_ord: int = 0
session: requests.Session = requests.Session()


class PlaceHolderSchema(BaseModel): ...


class DirectAgenticNodeTool(BaseTool):
    node_urls: list[str] = ["http://localhost:8000"]
    workflow: str = "research"
    name: str = ""
    description: str = ""
    args_schema: Type[BaseModel] = PlaceHolderSchema

    def __init__(self, node_urls: str | list[str], **kwargs) -> None:
        """DirectAgenticNodeTool constructor
        Keyword arguments:
        node_urls -- The URL of the node (str) or a list of URLs (list[str])
        workflow -- The workflow to run
        timeout -- The timeout for the job
        """
        try:
            if isinstance(node_urls, str):
                node_urls = [node_urls]
            args_schema, name, description = self.query_workflow(
                node_url=node_urls[0], workflow=kwargs["workflow"]
            )
            kwargs["node_urls"] = node_urls
            kwargs["name"] = name
            kwargs["description"] = description
            kwargs["args_schema"] = args_schema
        except Exception as e:
            log.error(f"Failed to properly initialize DirectAgenticNodeTool {e}")

        super().__init__(**kwargs)

    def _run(self, **kwargs) -> str:
        return self._do_run(**kwargs)

    @retry(stop=stop_after_attempt(3))
    def _do_run(self, **kwargs) -> str:
        global host_ord
        # Round robin selection of host
        host_ord = (host_ord + 1) % len(self.node_urls)
        # post the argument as request body to the node_url
        url = self.node_urls[host_ord]
        res = requests.post(f"{url}/workflow/{self.workflow}/run", json=kwargs)
        tid = res.text.strip('"')

        for _ in range(0, 30):
            res = requests.get(f"{url}/job/{tid}")
            job = Job.model_validate_json(res.text)
            if job.status != JobStatus.RUNNING:
                return job.output or ""
            time.sleep(1)
        log.error("Job took too long to complete")
        raise Exception("Job took too long to complete")

    @retry(stop=stop_after_delay(30), wait=wait_fixed(2))
    def query_workflow(
        self, node_url: str, workflow: str
    ) -> tuple[Type[BaseModel], str, str]:
        try:
            if isinstance(node_url, list):
                node_url = node_url[0]
            res = requests.get(f"{node_url}/node/description")
            if res.status_code != 200:
                raise Exception("Failed to get node description")
            wfs = json.loads(res.text)
            wf = Workflow(**wfs[workflow])

            fields = {k: Field(..., description=v) for k, v in wf.inputs.items()}
            fields["__annotations__"] = {k: str for k in wf.inputs.keys()}

            Schema = type(
                "AgenticToolSchema",
                (BaseModel,),
                fields,
            )

            return Schema, wf.name, wf.description
        except Exception as e:
            log.warning(f"Failed to query workflow: {e}")
            raise Exception(f"Failed to query workflow: {e}")


class RegistryAgenticNodeTool(BaseTool):
    node_ids: list[UUID] = [UUID("00000000-0000-0000-0000-000000000000")]
    workflow: str = "research"
    name: str = ""
    description: str = ""
    args_schema: Type[BaseModel] = PlaceHolderSchema

    def __init__(self, node_ids: UUID | list[UUID], **kwargs) -> None:
        """DirectAgenticNodeTool constructor
        Keyword arguments:
        node_ids -- The ID of the node (UUID) or a list of IDs (list[UUID])
        workflow -- The workflow to run
        timeout -- The timeout for the job
        When connecting to the registry it also requires following environment variables:
        REGISTRY_URL, BASIC_AUTH_USERNAME, BASIC_AUTH_PASSWORD
        """
        try:
            if settings.has_basic_auth():
                session.auth = (
                    settings.BASIC_AUTH_USERNAME,
                    settings.BASIC_AUTH_PASSWORD,
                )
            if isinstance(node_ids, UUID):
                node_ids = [node_ids]
            args_schema, name, description = self.query_workflow(
                node_id=str(node_ids[0]), workflow=kwargs["workflow"]
            )
            kwargs["node_ids"] = node_ids
            kwargs["name"] = name
            kwargs["description"] = description
            kwargs["args_schema"] = args_schema
        except Exception as e:
            log.error(f"Failed to properly initialize DirectAgenticNodeTool {e}")

        super().__init__(**kwargs)

    def _run(self, **kwargs) -> str:
        return self._do_run(**kwargs)

    @retry(stop=stop_after_attempt(3))
    def _do_run(self, **kwargs) -> str:
        global host_ord
        # Round robin selection of host
        host_ord = (host_ord + 1) % len(self.node_ids)
        # post the argument as request body to the node_url
        node_id = self.node_ids[host_ord]
        body = {"workflow": self.workflow, "inputs": kwargs, "node_id": str(node_id)}
        res = session.post(f"{settings.registry_rest_url()}/api/v1/jobs", json=body)
        if res.status_code != 200:
            raise Exception("Failed to start job")
        tid = res.text.strip('"')

        for _ in range(0, 30):
            res = session.get(f"{settings.registry_rest_url()}/api/v1/jobs/{tid}")
            job = res.json()
            if job["status"] != JobStatus.RUNNING:
                return job["result"] or ""
            time.sleep(1)
        log.error("Job took too long to complete")
        raise Exception("Job took too long to complete")

    @retry(stop=stop_after_delay(30), wait=wait_fixed(2))
    def query_workflow(
        self, node_id: str, workflow: str
    ) -> tuple[Type[BaseModel], str, str]:
        try:
            if isinstance(node_id, list):
                node_id = node_id[0]
            res = session.get(
                f"{settings.registry_rest_url()}/api/v1/nodes/{node_id}",
            )
            if res.status_code != 200:
                raise Exception(
                    f"Failed to fetch node description with status code {res.status_code}"
                )
            wfs = json.loads(res.text)["workflows"]
            wf = Workflow(**wfs[workflow])

            fields = {k: Field(..., description=v) for k, v in wf.inputs.items()}
            fields["__annotations__"] = {k: str for k in wf.inputs.keys()}

            Schema = type(
                "AgenticToolSchema",
                (BaseModel,),
                fields,
            )

            return Schema, wf.name, wf.description
        except Exception as e:
            log.warning(f"Failed to query workflow: {e}")
            raise Exception(f"Failed to query workflow: {e}")
